#include "rah-xmms-vis.h"
#include <xmms/plugin.h>

namespace RAHXMMSVIS {

void * xmalloc (size_t size) {
  void * ptr;
  ptr = malloc (size);
  if (!ptr) {
    exit (1);
  }
  return ptr;
}

void * xrealloc (void * mem, size_t size) {
  void * ptr;
  ptr = realloc (mem, size);
  if (!ptr) {
    exit (1);
  }
  return ptr;
}


string Misc::int2string (int i) {
  static uint size = 16;
  static char * s = NULL;
  int err = -1;

  if (!s) s = (char *) xmalloc (size * sizeof(char));

  while (err == -1) {
    err = snprintf (s, size, "%d", i);
    if (err == -1) {
      size += 16;
      s = (char *) xrealloc (s, size);
    }
  }

  return string (s);
}

int Misc::randomNumber (double lower, double upper) {
  return (int) ((lower-1)+(int) ((upper+1)*rand()/(RAND_MAX+1.0)));
}



void Node::connectPortRanges (list<Plug *>::iterator& plug,
                              list<Socket *>::iterator& socket,
                              unsigned count) {
  for (int i = 0;
       i < count;
       ++i, ++socket, ++plug) {
    cerr << i << endl;
    (*plug)->plugInto (*socket);
  }
}

bool Node::allSocketsHaveData () {
  for (list<Socket *>::iterator its = sockets.begin();
       its != sockets.end();
       ++its) {
    if (!(*its)->hasData()) return false;
  }
  return true;
}

void Plug::process () {
  cerr << DEBUG_STRING + "begin" <<  endl;
  if (_connectedSockets.empty()) {
    cerr << DEBUG_STRING + "no connected sockets, returning" <<  endl;
    return;
  }
  for (list<Socket *>::iterator its = _connectedSockets.begin();
       its != _connectedSockets.end(); 
       ++its) {
    if (!(*its)->node()->allSocketsHaveData()) {
      cerr << DEBUG_STRING + "the node connected from the plug doesn't have all its data, continuing with the rest" <<  endl;
      continue;
    }
    (*its)->process ();
  }
  cerr << DEBUG_STRING + "end" <<  endl;
}

void Socket::process () {
  cerr << DEBUG_STRING + "begin" <<  endl;
  if (!_node->allSocketsHaveData ()) {
    cerr << DEBUG_STRING + "this socket doesn't have all its data, returning" <<  endl;
    return;
  }
  _node->process ();
  cerr << DEBUG_STRING + "end" <<  endl;
}

void Socket::write (Data * data) {
  _dataBuffer.push_back (data);
}

Data * Socket::read () {
  Data * data = _dataBuffer.front ();
  _dataBuffer.pop_front();
  return data;
}

void Plug::write (Data * data) {
  cerr << DEBUG_STRING + "begin" <<  endl;
  if (_connectedSockets.empty()) {
    cerr << DEBUG_STRING + "no sockets connected, returning" << endl;
    return;
  }
  for (list<Socket *>::iterator its = _connectedSockets.begin();
       its != _connectedSockets.end ();
       ++its) {
  Data * nd = data->_type == Pulse ? (Data *) new PulseData ((PulseData *)data)
       : (Data *) new SurfaceData ( (const SDL_Surface *) ((SurfaceData *)data)->getSurface() );
  cerr << DEBUG_STRING + "writing data at " << data
       << " from this port, "
       << this
       << ", to port "
       << *its
       << endl;
    (*its)->write (nd);
  }
  delete data;
  cerr << DEBUG_STRING + "end" <<  endl;
}

void Plug::plugInto (Socket * socket) {
  if (socket->_type != _type) {
    string s (DEBUG_STRING + "port type mismatch");
    cerr << s << endl;
    throw Exception (s);
  }
  _connectedSockets.push_back (socket);
  socket->connect (this);
}

void Plug::unplugAllFrom (Socket * socket) {
  if (!socket) {
    string s (DEBUG_STRING + "socket is null");
    cerr << s << endl;
    throw Exception (s);
  }
  _connectedSockets.remove (socket);
  socket->unconnect ();
}





XmmsNode::XmmsNode () {
  int i;
  for (i = 0; i < 512; ++i) {
    plugs.push_back (new Plug (Pulse, this));
  }
}

void XmmsNode::process () {
  cerr << DEBUG_STRING + "begin" <<  endl;
  int i = 0;
  list<Plug *>::iterator itp;
  list<int>::iterator itd;
  for (itp = plugs.begin(),
       itd = _freqData.begin();
       itp != plugs.end();
       ++itp, ++itd, ++i) {
    cerr << DEBUG_STRING + "processing port " << i <<  endl;
    Data * data = new PulseData (*itd);
    (*itp)->write (data);
  }
  for (itp = plugs.begin ();
       itp != plugs.end ();
       ++itp) {
    (*itp)->process ();
  }
  cerr << DEBUG_STRING + "end" <<  endl;
}

void SDLNode::process () {
  cerr << DEBUG_STRING + "begin" <<  endl;
  list<SDL_Surface *>::iterator itsurf;
  list<Socket *>::iterator its;
  for (itsurf = _surfaces.begin(),
       its = sockets.begin();
       itsurf != _surfaces.end();
       ++itsurf, ++its) {
    SDL_Surface * surface = ((SurfaceData *) (*its)->read())->getSurface();
    SDL_BlitSurface (surface, NULL, *itsurf, NULL);
    delete surface;
    SDL_UpdateRect(*itsurf, 0, 0, (*itsurf)->w, (*itsurf)->h);
  }
  cerr << DEBUG_STRING + "end" <<  endl;
}

SDLNode::SDLNode () {
  int err;
  int i;
  SDL_Surface * surface;
  
  err = SDL_Init (SDL_INIT_VIDEO);
  if (err < 0) {
    string s (DEBUG_STRING + "couldn't initialise sdl, exiting");
    cerr << s << endl;
    throw Exception (s);
  }

  _numberOfScreens = 1;
  
  for (i = 0; i < _numberOfScreens; ++i) {
    surface = SDL_SetVideoMode(640, 480, 32, SDL_SWSURFACE);
    if (!surface) {
      cerr << DEBUG_STRING + "sdl screen " << i << " could not be initialised, continuing with the rest" << endl;
      continue;
    }
    _surfaces.push_back (surface);
  }
  
  if (_surfaces.size() == 0) {
    string s (DEBUG_STRING + "no sdl screens were initialised, exiting");
    cerr << s << endl;
    throw Exception (s);
  }

  for (i = 0; i < _numberOfScreens; ++i) {
    sockets.push_back (new Socket (Surface, this));
  }

}

void EQNode::process () {
  cerr << DEBUG_STRING + "begin" <<  endl;
  int band = 0;
  int val, i;
  bool doneTwice = false;
  bool doneOnce = false;
  PulseData * pulseData;
  list<Socket *>::iterator its = sockets.begin ();
  list<Plug *>::iterator itp = plugs.begin();
  for (band = 0;
       band < _bands && !doneTwice;
       ++band, ++itp) {
    cerr << DEBUG_STRING + "band: " << band <<  endl;
    val = 0;
    int numberOfData = (_inputs/2)/_bands;
    for (i = 0; i < numberOfData; ++i, ++its) {
      pulseData = (PulseData *) (*its)->read ();
      val += pulseData->getValue ();
      if (i != numberOfData-1) delete pulseData;
    }
    val /= numberOfData;
    pulseData->setValue (val); // overwrite old value
    (*itp)->write (pulseData);
    if ((band+1) == _bands) {
      if (doneOnce) {
        doneTwice = true;
      } else {
        doneOnce = true;
      }
      band = -1;
    }
  }
  for (itp = plugs.begin();
       itp != plugs.end ();
       ++itp) {
    (*itp)->process ();
  }
  cerr << DEBUG_STRING + "end" <<  endl;
}

void ImageFlickerNode::process () {
  cerr << DEBUG_STRING + "begin" <<  endl;
  list<Socket *>::iterator it = sockets.begin();
  PulseData * pulseData = (PulseData *) (*it)->read();
  int pulse = pulseData->getValue();
  ++it;
  SurfaceData * imageData = (SurfaceData *) (*it)->read();
  SDL_Surface * image = imageData->getSurface();
  ++it;
  SurfaceData * surfaceData = (SurfaceData *) (*it)->read();
  SDL_Surface * surface = surfaceData->getSurface();

//  SDL_SetAlpha (image, SDL_SRCALPHA|SDL_RLEACCEL, (pulse/65536)*256);
  SDL_SetAlpha (image, SDL_SRCALPHA|SDL_RLEACCEL, pulse);
  SDL_BlitSurface (image, NULL, surface, NULL);
  SDL_UpdateRect(surface, 0, 0, image->w, image->h);

  delete pulseData;
  delete imageData;
  plugs.front()->write (surfaceData);
  plugs.front()->process ();
  cerr << DEBUG_STRING + "end" <<  endl;
}

void ImageOverlayNode::process () {
  cerr << DEBUG_STRING + "begin" <<  endl;
  if (!allSocketsHaveData()) {
    cerr << DEBUG_STRING + "not all sockets have data, returning" <<  endl;
    return;
  }
  list<Socket *>::iterator its = sockets.begin();
  SurfaceData * background = (SurfaceData *) (*its)->read();
  SDL_Surface * bgsurface = background->getSurface();
  for (++its; its != sockets.end(); ++its) {
    PulseData * pulseData = (PulseData *) (*its)->read();
    int alpha = pulseData->getValue();
    alpha = (alpha/65536)*256;
    ++its;
    SDL_Surface * image = ((SurfaceData *)(*its)->read())->getSurface();
    SDL_SetAlpha (image, SDL_SRCALPHA|SDL_RLEACCEL, alpha);
    SDL_BlitSurface (image, NULL, bgsurface, NULL);
    SDL_UpdateRect(bgsurface, 0, 0, bgsurface->w, bgsurface->h);
    delete pulseData;
    delete image;
  }
  (plugs.front())->write (background);
  (plugs.front())->process ();
  cerr << DEBUG_STRING + "end" <<  endl;
}

} /* namespace RAHXMMSVIS */

using namespace RAHXMMSVIS;

Node * rootNode = 0;

list <Node *> baseNodes;

void start_plugin () {
  int eqbands = 3;
  Node * node;
  cerr << DEBUG_STRING + "begin" <<  endl;
  try {
    if (!rootNode) rootNode = new XmmsNode ();
    Node * eqnode = new EQNode (512, eqbands);
    list<Plug *>::iterator itp = rootNode->plugs.begin();
    list<Socket *>::iterator its = eqnode->sockets.begin();
    rootNode->connectPortRanges (itp, its, 512);
    Node * background = new BitmapNode (string("volcano.bmp"), 1);
    Node * image = new BitmapNode (string("leaf-plasma.bmp"), 1);
  //  for (i = 0; i < eqbands; ++i) {
      node = new ImageFlickerNode ();
      its = node->sockets.begin();
      itp = eqnode->plugs.begin();
      (eqnode->plugs.front())->plugInto (*its);
      its++;
      (image->plugs.front())->plugInto (*its);
      its++;
      (background->plugs.front())->plugInto (*its);
  //  }
    Node * sdlNode = new SDLNode ();
    node->plugs.front()->plugInto (sdlNode->sockets.front());
    baseNodes.push_back (rootNode);
    baseNodes.push_back (image);
    baseNodes.push_back (background);
  }
  catch (Exception e) {
    cerr << DEBUG_STRING + "caught exception: " + e._desc << endl;
    rootNode = 0;
  }
  cerr << DEBUG_STRING + "end" << endl;
}

void stop_plugin () {
  cerr << DEBUG_STRING + "begin" << endl;
  if (rootNode) {
    delete rootNode;
    rootNode = 0;
  }
  cerr << DEBUG_STRING + "end" << endl;
}

void render_freq (gint16 freq_data[2][256]) {
  cerr << DEBUG_STRING + "begin" <<  endl;
  if (!rootNode) return;
  list<int> data;
  int i;
  for (i = 0; i < 512; ++i) {
    data.push_back (freq_data[i<256?0:1][i<256?i:i-256]);
  }
  ((XmmsNode *)rootNode)->setData (data);
  for (list<Node *>::iterator itn = baseNodes.begin();
       itn != baseNodes.end();
       ++itn) {
    (*itn)->process ();
  }
  cerr << DEBUG_STRING + "end" <<  endl;
}
extern "C" {

VisPlugin rah_xmms_vis_plugin = {
  NULL,
  NULL,
  0,
  "RAH XMMS Visualisation Studio",
  0,
  256,
  start_plugin,
  stop_plugin,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  render_freq
};

VisPlugin *get_vplugin_info (void) {
  return &rah_xmms_vis_plugin;
}

} /* extern "C" */
